#! python3
# -*- coding: utf-8 -*-
"""Widget minitor
*** Inspired by wx.lib.eventwatcher ***

Author: Kazuya O'moto <komoto@jeol.co.jp>
"""
import warnings
import wx
from wx import aui
from wx import stc
import wx.lib.eventwatcher as ew
try:
    from framework import where
    from controls import CheckList
except ImportError:
    from .framework import where
    from .controls import CheckList


class EventMonitor(CheckList):
    """Event monitor with check-list

Args:
    parent : shellframe
    """
    parent = property(lambda self: self.__shellframe)
    target = property(lambda self: self.__watchedWidget)
    
    def __init__(self, parent, **kwargs):
        CheckList.__init__(self, parent,
                           style=wx.LC_REPORT|wx.LC_HRULES, **kwargs)
        
        self.__shellframe = parent
        self.__watchedWidget = None
        
        self.__dir = True # sort direction
        self.__items = []
        
        self.Font = wx.Font(9, wx.DEFAULT, wx.NORMAL, wx.NORMAL)
        
        self.alist = ( # assoc list of column names
            ("typeId",    62),
            ("typeName", 200),
            ("source",   200),
        )
        for k, (header, w) in enumerate(self.alist):
            self.InsertColumn(k, header, width=w)
        
        ## self.Bind(wx.EVT_MOTION, self.OnMotion)
        self.Bind(wx.EVT_LIST_COL_CLICK, self.OnSortItems)
        self.Bind(wx.EVT_LIST_ITEM_FOCUSED, self.OnItemFocused)
        ## self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.OnItemFocused)
        ## self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnItemActivated) # left-dclick, space
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnItemDClick)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)
        
        ## self.ew = ew
        self.add_module = ew.addModuleEvents
    
    def OnDestroy(self, evt):
        if evt.EventObject is self:
            self.unwatch()
        evt.Skip()
    
    ## --------------------------------
    ## EventWatcher wrapper interfaces 
    ## --------------------------------
    
    ew.buildWxEventMap() # build ew._eventBinders and ew._eventIdMap
    ew.addModuleEvents(aui) # + some additives
    ew.addModuleEvents(stc)
    
    ## Events that should not be watched by default
    ew._noWatchList = [
        wx.EVT_PAINT,
        wx.EVT_NC_PAINT,
        wx.EVT_ERASE_BACKGROUND,
        wx.EVT_IDLE,
        wx.EVT_UPDATE_UI,
        wx.EVT_UPDATE_UI_RANGE,
        wx.EVT_TOOL,
        wx.EVT_TOOL_RANGE, # menu items (typeId=10018)
        wx.EVT_MENU,
    ]
    
    @staticmethod
    def get_name(event):
        return ew._eventIdMap.get(event, 'Unknown')
    
    @staticmethod
    def get_binder(event):
        return next(x for x in ew._eventBinders if x.typeId == event)
    
    @staticmethod
    def get_watchlist():
        """All watched event binders except noWatchList"""
        return filter(lambda v: v not in ew._noWatchList,
                      ew._eventBinders)
    
    def get_actions(self, event, widget=None):
        """Wx.PyEventBinder and the handlers"""
        widget = widget or self.target
        if widget and hasattr(widget, '__event_handler__'):
            try:
                handlers = widget.__event_handler__[event]
                ## Exclude ew:onWatchedEvent by comparing names instead of objects
                ## return [a for a in handlers if a != self.onWatchedEvent]
                return [a for a in handlers if a.__name__ != 'onWatchedEvent']
            except KeyError:
                print("- No such event: {}".format(event))
    
    def watch(self, widget):
        """Begin watching"""
        if not widget:
            self.unwatch()
            return
        if not isinstance(widget, wx.Object):
            wx.MessageBox("Cannot watch the widget.\n\n"
                          "- {!r} is not a wx.Object.".format(widget))
            return
        self.unwatch()
        self.clear()
        self.__watchedWidget = widget
        ssmap = self.dump(widget, verbose=1)
        for binder in self.get_watchlist():
            widget.Bind(binder, self.onWatchedEvent)
            if binder.typeId in ssmap:
                self.append(binder.typeId)
        self.parent.handler("show_page", self)
        self.parent.handler("monitor_begin", self.target)
    
    def unwatch(self):
        """End watching"""
        if not self.target:
            return
        for binder in self.get_watchlist():
            if not self.target.Unbind(binder, handler=self.onWatchedEvent):
                print("- Failed to unbind {}:{}".format(binder.typeId, binder))
        self.parent.handler("monitor_end", self.target)
        self.__watchedWidget = None
    
    def onWatchedEvent(self, evt):
        if self:
            self.update(evt)
        evt.Skip()
    
    def dump(self, widget, verbose=True):
        """Dump all event handlers bound to the widget"""
        exclusions = [x.typeId for x in ew._noWatchList]
        ssmap = {}
        if not hasattr(widget, '__event_handler__'):
            return ssmap
        for event in sorted(widget.__event_handler__):
            actions = self.get_actions(event)
            if actions and event not in exclusions:
                ssmap[event] = actions
                if verbose:
                    name = self.get_name(event)
                    values = ('\n'+' '*41).join(where(a) for a in actions)
                    print("{:8d}:{:32s}{!s}".format(event, name, values))
        return ssmap
    
    ## --------------------------------
    ## Actions for event-logger items
    ## --------------------------------
    
    def update(self, evt):
        event = evt.EventType
        obj = evt.EventObject
        name = self.get_name(event)
        source = ew._makeSourceString(obj)
        
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', DeprecationWarning)
            attribs = ew._makeAttribString(evt)
        
        if wx.VERSION < (4,1,0): # ignore self insert
            if event == wx.EVT_LIST_INSERT_ITEM.typeId\
              and obj is self:
                return
        
        for i, item in enumerate(self.__items):
            if item[0] == event:
                item[1:] = [name, source, attribs]
                break
        else:
            i = len(self.__items)
            item = [event, name, source, attribs]
            self.__items.append(item)
            self.InsertItem(i, event)
        
        for j, v in enumerate(item[:-1]):
            self.SetItem(i, j, str(v))
        
        if i == self.FocusedItem:
            self.parent.handler("put_scratch", attribs)
        
        if self.IsItemChecked(i):
            actions = self.get_actions(evt.EventType)
            if actions:
                self.CheckItem(i, False)
                for f in actions:
                    self.parent.debugger.trace(f, evt)
        
        if self.GetItemBackgroundColour(i) != wx.Colour('yellow'):
            ## Don't run out of all timers and get warnings
            self.SetItemBackgroundColour(i, "yellow")
            def reset_color():
                if self and i < self.ItemCount:
                    self.SetItemBackgroundColour(i, 'white')
            wx.CallLater(1000, reset_color)
    
    def clear(self):
        self.DeleteAllItems()
        self.__items = []
    
    def append(self, event, bold=True):
        if event in (item[0] for item in self.__items):
            return
        
        i = len(self.__items)
        name = self.get_name(event)
        item = [event, name, '-', 'no data']
        self.__items.append(item)
        self.InsertItem(i, event)
        for j, v in enumerate(item[:-1]):
            self.SetItem(i, j, str(v))
        if bold:
            self.SetItemFont(i, self.Font.Bold())
    
    def OnSortItems(self, evt): #<wx._controls.ListEvent>
        n = self.ItemCount
        if n < 2:
            return
        lc = [self.__items[j] for j in range(n) if self.IsItemChecked(j)]
        ls = [self.__items[j] for j in range(n) if self.IsSelected(j)]
        f = self.__items[self.FocusedItem]
        
        col = evt.GetColumn()
        self.__dir = not self.__dir
        self.__items.sort(key=lambda v: v[col], reverse=self.__dir) # sort data
        
        for i, item in enumerate(self.__items):
            for j, v in enumerate(item[:-1]):
                self.SetItem(i, j, str(v))
            self.CheckItem(i, item in lc)  # check
            self.Select(i, item in ls)     # seleciton
            if self.get_actions(item[0]):
                self.SetItemFont(i, self.Font.Bold())
            else:
                self.SetItemFont(i, self.Font)
        self.Focus(self.__items.index(f))  # focus (one)
    
    ## def OnItemActivated(self, evt): #<wx._controls.ListEvent>
    ##     i = evt.Index
    ##     item = self.__items[i]
    ##     attribs = item[-1]
    ##     wx.CallAfter(wx.TipWindow, self, attribs, 512)
    ##     self.parent.handler("put_scratch", attribs)
    
    def OnItemFocused(self, evt): #<wx._controls.ListEvent>
        i = evt.Index
        item = self.__items[i]
        self.parent.handler("put_scratch", item[-1]) # attribs
        evt.Skip()
    
    def OnItemDClick(self, evt): #<wx._core.MouseEvent>
        i, flag = self.HitTest(evt.Position)
        if i >= 0:
            item = self.__items[i]
            wx.CallAfter(wx.TipWindow, self, item[-1], 512) # attribs
        evt.Skip()


if __name__ == "__main__":
    import mwx
    app = wx.App()
    frm = mwx.Frame(None)
    if 1:
        self = frm.shellframe
        frm.mon = EventMonitor(self)
        frm.mon.Show(0)
        self.Show()
        self.rootshell.write("self.mon.watch(self.mon)")
    frm.Show()
    app.MainLoop()
